import os
import shutil
import signal
import tarfile
import traceback
import uuid
from pathlib import Path
from typing import Dict, List, Set, Union

from ..common import jsonpickle
from ..config import config
from ..utils import (artifact_classes, assemble_path_object,
                     catalog_of_local_artifact, convert_dflow_list, copy_file,
                     expand, flatten, randstr, remove_empty_dir_tag)
from .opio import (Artifact, BigParameter, HDF5Dataset, HDF5Datasets,
                   NestedDict, Parameter)


def get_slices(path_object, slices):
    if isinstance(slices, int):
        return path_object[slices]
    elif isinstance(slices, str):
        tmp = path_object
        for f in slices.split("."):
            if isinstance(tmp, dict):
                tmp = tmp[f]
            elif isinstance(tmp, list):
                tmp = tmp[int(f)]
        return tmp
    elif isinstance(slices, list):
        return [get_slices(path_object, i) for i in slices]
    return path_object


def handle_input_artifact(name, sign, slices=None, data_root="/tmp",
                          sub_path=None, n_parts=None, keys_of_parts=None,
                          path=None, prefix=None):
    root = None
    if n_parts is not None:
        path_object = []
        for i in range(n_parts):
            art_path = '%s/inputs/artifacts/dflow_%s_%s' % (data_root, name, i)
            if config["detect_empty_dir"]:
                remove_empty_dir_tag(art_path)
            po = assemble_path_object(art_path)
            if prefix is not None:
                po = get_slices(po, prefix[i])
            if not po:
                path_object.append(art_path)
            elif isinstance(po, list):
                path_object += po
            else:
                path_object.append(po)
    elif keys_of_parts is not None:
        path_object = {}
        for i in keys_of_parts:
            art_path = '%s/inputs/artifacts/dflow_%s_%s' % (data_root, name, i)
            if config["detect_empty_dir"]:
                remove_empty_dir_tag(art_path)
            po = assemble_path_object(art_path)
            if prefix is not None:
                po = get_slices(po, prefix[i])
            if not po:
                path_object[i] = art_path
            else:
                path_object[i] = po
        path_object = expand(path_object)
    else:
        art_path = '%s/inputs/artifacts/%s' % (data_root, name) \
            if path is None else path
        if sub_path is not None:
            art_path = os.path.join(art_path, sub_path)
        root = art_path
        if not os.path.exists(art_path):  # for optional artifact
            return None
        if config["detect_empty_dir"]:
            remove_empty_dir_tag(art_path)
        path_object = assemble_path_object(art_path)
        path_object = get_slices(path_object, prefix)

    sign_type = sign.type
    if getattr(sign_type, "__origin__", None) == Union:
        args = sign_type.__args__
        if HDF5Datasets in args:
            if isinstance(path_object, list) and len(path_object) > 0 and all([
                isinstance(p, str) and p.endswith(".h5")
                    for p in path_object]):
                sign_type = HDF5Datasets
            elif art_path.endswith(".h5"):
                sign_type = HDF5Datasets
            elif args[0] == HDF5Datasets:
                sign_type = args[1]
            elif args[1] == HDF5Datasets:
                sign_type = args[0]

    if sign_type == HDF5Datasets:
        import h5py
        if os.path.isfile(art_path):
            path_object = [art_path]
        assert isinstance(path_object, list)
        res = []
        for path in path_object:
            f = h5py.File(path, "r")
            datasets = {}
            for k in f.keys():
                d = HDF5Dataset(f, k)
                datasets[k] = None if d.is_none() else d
            if set(datasets.keys()) == {str(i) for i in range(len(datasets))} \
                    and isinstance(res, list):
                # concat when all datasets are lists
                res += expand(datasets)
            else:
                # merge otherwise
                if isinstance(res, list):
                    res = flatten(res)
                res.update(datasets)

        if isinstance(res, dict):
            res = expand(res)
        res = get_slices(res, slices)
    else:
        path_object = get_slices(path_object, slices)

    if sign_type in [str, Path]:
        if path_object is None or isinstance(path_object, str):
            res = path_object
        elif isinstance(path_object, list) and len(path_object) == 1 and (
            path_object[0] is None or isinstance(path_object[0], str)) \
                and sign.sub_path:
            res = path_object[0]
        else:
            res = art_path
        res = path_or_none(res) if sign_type == Path else res
    elif sign_type in [List[str], List[Path], Set[str], Set[Path]]:
        if path_object is None:
            return None
        elif isinstance(path_object, str):
            res = [path_object]
        elif isinstance(path_object, list) and all([
                p is None or isinstance(p, str) for p in path_object]):
            res = path_object
        else:
            res = list(flatten(path_object).values())

        if sign_type == List[str]:
            pass
        elif sign_type == List[Path]:
            res = path_or_none(res)
        elif sign_type == Set[str]:
            res = set(res)
        else:
            res = set(path_or_none(res))
    elif sign_type in [Dict[str, str], NestedDict[str]]:
        res = path_object
    elif sign_type in [Dict[str, Path], NestedDict[Path]]:
        res = path_or_none(path_object)

    if res is None:
        return None

    _cls = res.__class__
    res = artifact_classes.get(_cls, lambda x: x)(res)
    res.art_root = root
    return res


def path_or_none(p):
    if p is None:
        return None
    elif isinstance(p, list):
        return [path_or_none(i) for i in p]
    elif isinstance(p, dict):
        return {k: path_or_none(v) for k, v in p.items()}
    else:
        return Path(p)


def handle_input_parameter(name, value, sign, slices=None, data_root="/tmp"):
    if "dflow_list_item" in value:
        dflow_list = []
        for item in jsonpickle.loads(value):
            dflow_list += jsonpickle.loads(item)
        obj = convert_dflow_list(dflow_list)
    elif isinstance(sign, BigParameter) and config["mode"] != "debug":
        with open(data_root + "/inputs/parameters/" + name, "r") as f:
            content = jsonpickle.loads(f.read())
            obj = content
    else:
        if isinstance(sign, (Parameter, BigParameter)):
            sign = sign.type
        if sign == str and slices is None:
            obj = value
        else:
            obj = jsonpickle.loads(value)

    if slices is not None:
        assert isinstance(
            obj, list), "Only parameters of type list can be sliced, while %s"\
            " is not list" % obj
        if isinstance(slices, list):
            obj = [obj[i] for i in slices]
        else:
            obj = obj[slices]

    return obj


def slice_to_dir(slice):
    return str(slice).replace(".", "/")


def handle_output_artifact(name, value, sign, slices=None, data_root="/tmp",
                           create_dir=False, symlink=False):
    if os.path.isdir(data_root + '/outputs/artifacts/' + name):
        shutil.rmtree(data_root + '/outputs/artifacts/' + name)
    path_list = []
    if sign.type == HDF5Datasets:
        import h5py
        os.makedirs(data_root + '/outputs/artifacts/' + name, exist_ok=True)
        h5_name = "%s.h5" % uuid.uuid4()
        h5_path = '%s/outputs/artifacts/%s/%s' % (data_root, name, h5_name)
        if isinstance(slices, list):
            # merge lists
            assert isinstance(value, list) and len(slices) == len(value)
            items = [(str(s), v) for s, v in zip(slices, value)]
            slices = 0
        else:
            items = flatten(value).items()
        with h5py.File(h5_path, "w") as f:
            for s, v in items:
                if isinstance(v, Path):
                    if v.is_file():
                        try:
                            data = v.read_text(encoding="utf-8")
                            dtype = "utf-8"
                        except Exception:
                            import numpy as np
                            data = np.void(v.read_bytes())
                            dtype = "binary"
                        d = f.create_dataset(s, data=data)
                        d.attrs["type"] = "file"
                        d.attrs["path"] = str(v)
                        d.attrs["dtype"] = dtype
                    elif v.is_dir():
                        tgz_path = Path("%s.tgz" % v)
                        tf = tarfile.open(tgz_path, "w:gz", dereference=True)
                        tf.add(v)
                        tf.close()
                        import numpy as np
                        d = f.create_dataset(s, data=np.void(
                            tgz_path.read_bytes()))
                        d.attrs["type"] = "dir"
                        d.attrs["path"] = str(v)
                        d.attrs["dtype"] = "binary"
                elif isinstance(v, HDF5Dataset):
                    d = f.create_dataset(s, data=v.dataset[()])
                    d.attrs.update(v.dataset.attrs)
                elif v is None:
                    d = f.create_dataset(s, data="")
                    d.attrs["type"] = "null"
                else:
                    d = f.create_dataset(s, data=v)
                    d.attrs["type"] = "data"
                    if isinstance(v, str):
                        d.attrs["dtype"] = "utf-8"
        path_list.append({"dflow_list_item": h5_name, "order": slices or 0})
    if sign.type in [str, Path]:
        os.makedirs(data_root + '/outputs/artifacts/' + name, exist_ok=True)
        if slices is None:
            slices = 0
        path_list.append(copy_results_and_return_path_item(
            value, name, slices, data_root,
            slice_to_dir(slices) if create_dir else None, symlink=symlink))
    elif sign.type in [List[str], List[Path], Set[str], Set[Path]]:
        os.makedirs(data_root + '/outputs/artifacts/' + name, exist_ok=True)
        if slices is not None:
            if isinstance(slices, int):
                for path in value:
                    path_list.append(copy_results_and_return_path_item(
                        path, name, slices, data_root,
                        slice_to_dir(slices) if create_dir else None,
                        symlink=symlink))
            else:
                assert len(slices) == len(value)
                for path, s in zip(value, slices):
                    if isinstance(path, list):
                        for p in path:
                            path_list.append(
                                copy_results_and_return_path_item(
                                    p, name, s, data_root, slice_to_dir(
                                        s) if create_dir else None,
                                    symlink=symlink))
                    else:
                        path_list.append(copy_results_and_return_path_item(
                            path, name, s, data_root, slice_to_dir(
                                s) if create_dir else None, symlink=symlink))
        else:
            for s, path in enumerate(value):
                path_list.append(copy_results_and_return_path_item(
                    path, name, s, data_root, symlink=symlink))
    elif sign.type in [Dict[str, str], Dict[str, Path]]:
        os.makedirs(data_root + '/outputs/artifacts/' + name, exist_ok=True)
        for s, path in value.items():
            path_list.append(copy_results_and_return_path_item(
                path, name, s, data_root, symlink=symlink))
    elif sign.type in [NestedDict[str], NestedDict[Path]]:
        os.makedirs(data_root + '/outputs/artifacts/' + name, exist_ok=True)
        for s, path in flatten(value).items():
            path_list.append(copy_results_and_return_path_item(
                path, name, s, data_root, symlink=symlink))

    os.makedirs(data_root + "/outputs/artifacts/%s/%s" % (name, config[
        "catalog_dir_name"]), exist_ok=True)
    with open(data_root + "/outputs/artifacts/%s/%s/%s" % (name, config[
            "catalog_dir_name"], uuid.uuid4()), "w") as f:
        f.write(jsonpickle.dumps({"path_list": path_list}))
    if config["detect_empty_dir"]:
        handle_empty_dir(data_root + "/outputs/artifacts/%s" % name)
    if config["save_path_as_parameter"]:
        with open(data_root + '/outputs/parameters/dflow_%s_path_list'
                  % name, 'w') as f:
            f.write(jsonpickle.dumps(path_list))


def handle_output_parameter(name, value, sign, slices=None, data_root="/tmp"):
    if slices is not None:
        if isinstance(slices, list):
            assert isinstance(value, list) and len(slices) == len(value)
            res = [{"dflow_list_item": v, "order": s}
                   for v, s in zip(value, slices)]
        else:
            res = [{"dflow_list_item": value, "order": slices}]
        with open(data_root + '/outputs/parameters/' + name, 'w') as f:
            f.write(jsonpickle.dumps(res))
    elif isinstance(sign, BigParameter) and config["mode"] != "debug":
        with open(data_root + "/outputs/parameters/" + name, "w") as f:
            f.write(jsonpickle.dumps(value))
    else:
        if isinstance(sign, (Parameter, BigParameter)):
            sign = sign.type
        if sign == str:
            with open(data_root + '/outputs/parameters/' + name, 'w') as f:
                f.write(value)
        else:
            with open(data_root + '/outputs/parameters/' + name, 'w') as f:
                f.write(jsonpickle.dumps(value))


def copy_results_and_return_path_item(path, name, order, data_root="/tmp",
                                      slice_dir=None, symlink=False):
    if (path and os.path.exists(str(path))) or symlink:
        return {"dflow_list_item": copy_results(
                    path, name, data_root, slice_dir, symlink), "order": order}
    else:
        return {"dflow_list_item": None, "order": order}


def copy_results(source, name, data_root="/tmp", slice_dir=None,
                 symlink=False):
    source = str(source)
    # if refer to input artifact
    if source.find(data_root + "/inputs/artifacts/") == 0:
        # retain original directory structure
        i = source.find("/", len(data_root + "/inputs/artifacts/"))
        if i == -1:
            rel_path = randstr()
        else:
            rel_path = source[i+1:]
        if slice_dir is not None:
            rel_path = "%s/%s" % (slice_dir, rel_path)
        target = data_root + "/outputs/artifacts/%s/%s" % (name, rel_path)
        if symlink:
            os.makedirs(os.path.abspath(os.path.dirname(target)),
                        exist_ok=True)
            os.symlink(os.path.abspath(source), target)
        else:
            copy_file(source, target, shutil.copy)
        if rel_path[:1] == "/":
            rel_path = rel_path[1:]
        return rel_path
    else:
        cwd = os.getcwd()
        if not cwd.endswith("/"):
            cwd = cwd + "/"
        if source.startswith(cwd):
            source = source[len(cwd):]
        rel_path = source[1:] if source[:1] == "/" else source
        if slice_dir is not None:
            rel_path = "%s/%s" % (slice_dir, rel_path)
        target = data_root + "/outputs/artifacts/%s/%s" % (name, rel_path)
        if symlink:
            os.makedirs(os.path.abspath(os.path.dirname(target)),
                        exist_ok=True)
            os.symlink(os.path.abspath(source), target)
        else:
            copy_file(source, target)
        return rel_path


def handle_empty_dir(path):
    # touch an empty file in each empty dir, as object storage will ignore
    # empty dirs
    for dn, ds, fs in os.walk(path, followlinks=True):
        if len(ds) == 0 and len(fs) == 0:
            with open(os.path.join(dn, ".empty_dir"), "w"):
                pass


def handle_lineage(wf_name, pod_name, op_obj, input_urns, workflow_urn,
                   data_root="/tmp"):
    task_name = wf_name + "/" + pod_name
    output_sign = op_obj.get_output_sign()
    output_uris = {}
    for name, sign in output_sign.items():
        if isinstance(sign, Artifact):
            output_uris[name] = op_obj.get_output_artifact_storage_key(name)
    input_urns = {name: jsonpickle.loads(urn) if urn[:1] == "["
                  else urn for name, urn in input_urns.items()}
    output_urns = config["lineage"].register_task(
        task_name, input_urns, output_uris, workflow_urn)
    for name, urn in output_urns.items():
        with open("%s/outputs/parameters/dflow_%s_urn" % (data_root, name),
                  "w") as f:
            f.write(urn)


def absolutize(path):
    if path is None:
        return None
    if isinstance(path, str):
        return os.path.abspath(path)
    if isinstance(path, Path):
        return path.absolute()
    if isinstance(path, list):
        return [absolutize(p) for p in path]
    if isinstance(path, dict):
        return {k: absolutize(p) for k, p in path.items()}


def absolutize_hdf5(obj):
    if isinstance(obj, Path):
        return obj.absolute()
    if isinstance(obj, list):
        return [absolutize(p) for p in obj]
    if isinstance(obj, dict):
        return {k: absolutize(p) for k, p in obj.items()}
    return obj


def sigalrm_handler(signum, frame):
    raise TimeoutError("Timeout")


def try_to_execute(input, slice_dir, op_obj, output_sign, cwd, timeout=None):
    os.chdir(cwd)
    if slice_dir is not None:
        os.makedirs(slice_dir, exist_ok=True)
        os.chdir(slice_dir)
    if timeout is not None:
        signal.signal(signal.SIGALRM, sigalrm_handler)
        signal.alarm(timeout)
    try:
        output = op_obj.execute(input)
        for n, s in output_sign.items():
            if isinstance(s, Artifact) and s.type != HDF5Datasets:
                output[n] = absolutize(output[n])
            elif isinstance(s, Artifact) and s.type == HDF5Datasets:
                output[n] = absolutize_hdf5(output[n])
        os.chdir(cwd)
        return output, None
    except Exception as e:
        traceback.print_exc()
        os.chdir(cwd)
        if op_obj.outputs:
            return op_obj.outputs, e
        else:
            return None, e
    finally:
        if timeout is not None:
            signal.alarm(0)


def get_input_slices(name, data_root="/tmp"):
    art_path = '%s/inputs/artifacts/%s' % (data_root, name)
    catalog = catalog_of_local_artifact(art_path)
    slices = [item["order"] for item in catalog]
    slices.sort()
    return slices



import warnings
from typing import Any, List, Union

dflow = Any


class _Context(object):
    """Global Context Manager."""

    def __init__(self):
        """Init context with default to false."""
        self._in_context = False
        self.current_workflow = None

    @property
    def in_context(self) -> bool:
        """whether it is in context environment or not."""
        return self._in_context

    @in_context.setter
    def in_context(self, value: bool):
        if not isinstance(value, bool):
            raise ValueError(f'Unsupported in_context value: {value}.'
                             f'Expected bool, got {type(value)}')
        if not value:
            self._reset()
        self._in_context = value

    def _reset(self):
        """Reset context to origin condition.

        (for context exit usage only.)
        """
        self._in_context = False
        self.current_workflow = None

    def to_in_context(self):
        """Switch to be in context environment."""
        if self.in_context:
            warnings.warning(
                'Already in context. But call `to_in_context()` again.')
        self._in_context = True

    def to_out_context(self):
        """Switch to be not in context environment."""
        if not self.in_context:
            warnings.warning(
                'Not in context. But call `to_out_context()` again.')
        self._in_context = False

    def registry_step(self,
                      step: Union['dflow.Step',
                                  List['dflow.Step'],
                                  'dflow.Task',
                                  List['dflow.Task']]):
        """Registry step to context."""
        self.current_workflow.add(step)


GLOBAL_CONTEXT = _Context()


class Range_Context(object):
    """Local context for range."""

    def __init__(self):
        """Init context with default to false."""
        self.range_param_name = None
        self.range_target_name = None
        self._in_context = False
        self.current_step = None
        self.range_param_len = 0

    @property
    def in_context(self) -> bool:
        """whether it is in context environment or not."""
        return self._in_context

    @in_context.setter
    def in_context(self, value: bool):
        if not isinstance(value, bool):
            raise ValueError(f'Unsupported in_context value: {value}.'
                             f'Expected bool, got {type(value)}')
        if not value:
            self._reset()
        self._in_context = value

    def _reset(self):
        """Reset context to origin condition.

        (for context exit usage only.)
        """
        self._in_context = False
        self.current_workflow = None
        self.range_param_name = None
        self.range_target_name = None
        self.range_param_len = 0

    def to_in_context(self):
        """Switch to be in context environment."""
        if self.in_context:
            warnings.warning(
                'Already in context. But call `to_in_context()` again.')
        self._in_context = True

    def to_out_context(self):
        """Switch to be not in context environment."""
        if not self.in_context:
            warnings.warning(
                'Not in context. But call `to_out_context()` again.')
        self._in_context = False

    def get_current_range_param_name(self):
        if self.range_param_name:
            return self.range_param_name
        else:
            raise SyntaxError("It seems not right.")


GLOBAL_RANGE_CONTEXT = Range_Context()



import inspect
import os
import shutil
from typing import Union

from dp.launching.cli import to_runner
from dp.launching.typing import BaseModel, Field
from dp.launching.typing.io import InputFilePath, OutputDirectory
from pydantic.main import ModelMetaclass

from dflow import (OPTemplate, S3Artifact, Step, Workflow, download_artifact,
                   upload_s3)
from dflow.python import OP, Artifact, BigParameter, Parameter
from dflow.python.utils import (handle_input_artifact, handle_output_artifact,
                                handle_output_parameter)


def OP_template_to_parser(templ: OPTemplate, version: str = "1.0.0"):
    ns = {"output_dir": Field(default='./outputs', description='output dir'),
          "__annotations__": {"output_dir": OutputDirectory}}
    for name, par in templ.inputs.parameters.items():
        ns["__annotations__"][name] = par.type
        kwargs = {}
        if hasattr(par, "description"):
            kwargs["description"] = par.description
        if hasattr(par, "value"):
            kwargs["default"] = par.value
        ns[name] = Field(**kwargs)
    for name, art in templ.inputs.artifacts.items():
        ns["__annotations__"][name] = InputFilePath
        kwargs = {}
        if hasattr(art, "description"):
            kwargs["description"] = art.description
        ns[name] = Field(**kwargs)

    OPOptions = ModelMetaclass.__new__(ModelMetaclass, "OPOptions",
                                       (BaseModel,), ns)

    def op_runner(opts: OPOptions) -> int:
        step = Step("dflow-main", templ, parameters={
            name: getattr(opts, name)
            for name in templ.inputs.parameters.keys()
        }, artifacts={
            name: S3Artifact(key=upload_s3(getattr(opts, name).get_path()))
            for name in templ.inputs.artifacts.keys()
        })
        wf = Workflow(templ.name)
        wf.add(step)
        wf.submit()
        wf.wait()
        assert wf.query_status() == "Succeeded"
        step = wf.query_step("dflow-main")[0]
        if hasattr(step, "outputs") and hasattr(step.outputs, "parameters"):
            os.makedirs('outputs/parameters', exist_ok=True)
            for name, par in step.outputs.parameters.items():
                with open('outputs/parameters/%s' % name, 'w') as f:
                    f.write(par["value"])
        if hasattr(step, "outputs") and hasattr(step.outputs, "artifacts"):
            for name, art in step.outputs.artifacts.items():
                download_artifact(art, path='outputs/artifact/%s' % name,
                                  remove_catalog=False)
        return 0

    def to_parser():
        return to_runner(
            OPOptions,
            op_runner,
            version=version,
        )
    return to_parser


def python_OP_to_parser(op: OP, version: str = "1.0.0"):
    ns = {"output_dir": Field(default='./outputs', description='output dir'),
          "__annotations__": {"output_dir": OutputDirectory}}
    for name, sign in op.get_input_sign().items():
        if isinstance(sign, Artifact):
            ns["__annotations__"][name] = InputFilePath
            kwargs = {}
            if hasattr(sign, "description"):
                kwargs["description"] = sign.description
            ns[name] = Field(**kwargs)
        elif isinstance(sign, (Parameter, BigParameter)):
            ns["__annotations__"][name] = sign.type
            kwargs = {}
            if hasattr(sign, "description"):
                kwargs["description"] = sign.description
            if hasattr(sign, "default"):
                kwargs["default"] = sign.default
            ns[name] = Field(**kwargs)
        else:
            ns["__annotations__"][name] = sign
            ns[name] = Field()

    OPOptions = ModelMetaclass.__new__(ModelMetaclass, "OPOptions",
                                       (BaseModel,), ns)

    def op_runner(opts: OPOptions) -> int:
        op_in = {}
        for name, sign in op.get_input_sign().items():
            if isinstance(sign, Artifact):
                op_in[name] = handle_input_artifact(
                    name, sign,
                    path=os.path.abspath(getattr(opts, name).get_path()))
            else:
                op_in[name] = getattr(opts, name)
        op_out = op.execute(op_in)
        os.makedirs('outputs/parameters', exist_ok=True)
        os.makedirs('outputs/artifacts', exist_ok=True)
        for name, sign in op.get_output_sign().items():
            value = op_out[name]
            if isinstance(sign, Artifact):
                if os.path.isdir('outputs/artifacts/%s' % name):
                    shutil.rmtree('outputs/artifacts/%s' % name)
                handle_output_artifact(name, value, sign, data_root=".")
            else:
                handle_output_parameter(name, value, sign, data_root=".")
        return 0

    def to_parser():
        return to_runner(
            OPOptions,
            op_runner,
            version=version,
        )
    return to_parser


def OP_to_parser(op: Union[OP, OPTemplate], version: str = "1.0.0"):
    if inspect.isclass(op) and issubclass(op, OP):
        return python_OP_to_parser(op(), version)
    elif isinstance(op, OP):
        return python_OP_to_parser(op, version)
    elif isinstance(op, OPTemplate):
        return OP_template_to_parser(op, version)
    else:
        raise ValueError("Only Python OP/OP template supported")



import os
from typing import Optional

import oss2

from ..config import s3_config
from ..utils import StorageClient


class OSSClient(StorageClient):
    def __init__(
            self,
            endpoint: Optional[str] = None,
            bucket_name: Optional[str] = None,
            access_key_id: Optional[str] = None,
            access_key_secret: Optional[str] = None,
    ) -> None:
        if endpoint is None:
            endpoint = os.environ.get("OSS_ENDPOINT")
        if bucket_name is None:
            bucket_name = os.environ.get("OSS_BUCKET_NAME")
        if access_key_id is None:
            access_key_id = os.environ.get("OSS_ACCESS_KEY_ID")
        if access_key_secret is None:
            access_key_secret = os.environ.get("OSS_ACCESS_KEY_SECRET")
        self.endpoint = endpoint
        self.bucket_name = bucket_name
        self.access_key_id = access_key_id
        self.access_key_secret = access_key_secret
        auth = oss2.Auth(access_key_id, access_key_secret)
        bucket = oss2.Bucket(auth, endpoint, bucket_name)
        self.bucket = bucket

    def to_dict(self):
        retained_keys = ["endpoint", "bucket_name"]
        return {k: self.__dict__[k] for k in retained_keys}

    def prefixing(self, key):
        if not key.startswith(s3_config["repo_prefix"]):
            return s3_config["repo_prefix"] + key
        return key

    def unprefixing(self, key):
        if key.startswith(s3_config["repo_prefix"]):
            return key[len(s3_config["repo_prefix"]):]
        return key

    def upload(self, key, path):
        self.bucket.put_object_from_file(self.prefixing(key), path)

    def download(self, key, path):
        if os.path.dirname(path):
            os.makedirs(os.path.dirname(path), exist_ok=True)
        self.bucket.get_object_to_file(self.prefixing(key), path)

    def list(self, prefix, recursive=False):
        prefix = self.prefixing(prefix)
        keys = []
        if recursive:
            marker = ""
            while True:
                r = self.bucket.list_objects(prefix, marker=marker)
                for obj in r.object_list:
                    if not obj.key.endswith("/"):
                        keys.append(self.unprefixing(obj.key))
                if not r.is_truncated:
                    break
                marker = r.next_marker
        else:
            marker = ""
            while True:
                r = self.bucket.list_objects(prefix, delimiter="/",
                                             marker=marker)
                for obj in r.object_list:
                    if obj.key == prefix and obj.key.endswith("/"):
                        continue
                    keys.append(self.unprefixing(obj.key))
                for key in r.prefix_list:
                    keys.append(self.unprefixing(key))
                if not r.is_truncated:
                    break
                marker = r.next_marker
        return keys

    def copy(self, src, dst):
        self.bucket.copy_object(self.bucket_name, self.prefixing(src),
                                self.prefixing(dst))

    def get_md5(self, key):
        return self.bucket.get_object_meta(self.prefixing(key).etag)



import os
from typing import Dict, List, Optional, Union

from dp.metadata import Dataset, MetadataContext
from dp.metadata.entity.task import Task
from dp.metadata.entity.workflow import WorkFlow

from .. import LineageClient

config = {
    "gms_endpoint": os.environ.get("METADATA_GMS_ENDPOINT",
                                   "https://datahub-gms.dp.tech"),
    "project": os.environ.get("METADATA_PROJECT", None),
    "token": os.environ.get("METADATA_TOKEN", None),
}


class MetadataClient(LineageClient):
    def __init__(self, gms_endpoint=None, project=None, token=None):
        if gms_endpoint is None:
            gms_endpoint = config["gms_endpoint"]
        if project is None:
            project = config["project"]
        if token is None:
            token = config["token"]
        self.gms_endpoint = gms_endpoint
        self.project = project
        self.token = token

    def register_workflow(
            self,
            workflow_name: str) -> str:
        with MetadataContext(project=self.project, endpoint=self.gms_endpoint,
                             token=self.token) as context:
            client = context.client
            workflow = WorkFlow(workflow_name)
            job = client.prepare_workflow(workflow)
            client.begin_workflow(job)
        return str(job.urn)

    def register_artifact(
            self,
            namespace: str,
            name: str,
            uri: str,
            description: str = "",
            tags: Optional[List[str]] = None,
            properties: Optional[Dict[str, str]] = None,
            **kwargs) -> str:
        if tags is None:
            tags = []
        if properties is None:
            properties = {}
        with MetadataContext(project=self.project, endpoint=self.gms_endpoint,
                             token=self.token) as context:
            client = context.client
            urn = Dataset.gen_urn(context, namespace, name)
            ds = Dataset(urn=urn, display_name=name, uri=uri,
                         description=description, tags=tags,
                         properties=properties)
            client.update_dataset(ds)
        return urn

    def register_task(
            self,
            task_name: str,
            input_urns: Dict[str, Union[str, List[str]]],
            output_uris: Dict[str, str],
            workflow_urn: str) -> Dict[str, str]:
        with MetadataContext(project=self.project, endpoint=self.gms_endpoint,
                             token=self.token) as context:
            client = context.client
            task = Task(task_name, workflow_urn)
            job = client.prepare_job(task)
            inputs = []
            for urn in input_urns.values():
                if isinstance(urn, list):
                    inputs += urn
                elif isinstance(urn, str):
                    inputs.append(urn)
            inputs = list(filter(lambda x: x != "", inputs))
            run = client.begin_job(job, inputs=inputs)
            output_urns = {}
            for name, uri in output_uris.items():
                urn = Dataset.gen_urn(context, task_name, name)
                ds = Dataset(urn=urn, display_name=name, uri=uri)
                client.update_dataset(ds)
                output_urns[name] = urn
            client.end_job(run, outputs=list(output_urns.values()))
        return output_urns

    def get_artifact_metadata(self, urn: str) -> object:
        with MetadataContext(project=self.project, endpoint=self.gms_endpoint,
                             token=self.token) as context:
            client = context.client
            ds = client.get_dataset(urn)
        return ds



import abc
import json
import os
import re
from abc import ABC
from copy import deepcopy
from getpass import getpass
from typing import Dict, List, Optional, Union

from ..common import CustomArtifact, S3Artifact
from ..config import config
from ..executor import Executor, render_script_with_tmp_root, run_script
from ..io import InputArtifact, InputParameter, OutputParameter
from ..op_template import ScriptOPTemplate
from ..python import PythonOPTemplate
from ..python.python_op_template import handle_packages_script
from ..util_ops import InitArtifactForSlices
from ..utils import randstr, upload_s3

try:
    from argo.workflows.client import (V1HostPathVolumeSource, V1Volume,
                                       V1VolumeMount)
except Exception:
    pass


def update_dict(d1: dict, d2: dict) -> None:
    for k, v in d2.items():
        if isinstance(v, dict) and k in d1 and isinstance(d1[k], dict):
            update_dict(d1[k], v)
        else:
            d1[k] = v


class DispatcherExecutor(Executor):
    """
    Dispatcher executor

    Args:
        host: remote host
        queue_name: queue name
        port: SSH port
        username: username
        password: password
        private_key_file: private key file for SSH
        image: image for dispatcher
        image_pull_policy: image pull policy for dispatcher
        command: command for dispatcher
        remote_command: command for running the script remotely
        map_tmp_dir: map /tmp to ./tmp
        machine_dict: machine config for dispatcher
        resources_dict: resources config for dispatcher
        task_dict: task config for dispatcher
        json_file: JSON file containing machine and resources config
        docker_executable: docker executable to run remotely
        singularity_executable: singularity executable to run remotely
        podman_executable: podman executable to run remotely
        remote_root: remote root path for working
        retry_on_submission_error: max retries on submission error
        merge_sliced_step: handle multi slices in a single dispatcher job
        terminate_grace_period: grace period in seconds after termination
            for collecting outputs
    """

    def __init__(self,
                 host: Optional[str] = None,
                 queue_name: Optional[str] = None,
                 port: int = 22,
                 username: str = "root",
                 password: Optional[str] = None,
                 private_key_file: Optional[os.PathLike] = None,
                 image: Optional[str] = None,
                 image_pull_policy: Optional[str] = None,
                 command: Union[str, List[str]] = "python3",
                 remote_command: Union[str, List[str]] = None,
                 map_tmp_dir: bool = True,
                 machine_dict: Optional[dict] = None,
                 resources_dict: Optional[dict] = None,
                 task_dict: Optional[dict] = None,
                 json_file: Optional[os.PathLike] = None,
                 docker_executable: Optional[str] = None,
                 singularity_executable: Optional[str] = None,
                 container_args: str = "",
                 podman_executable: Optional[str] = None,
                 remote_root: Optional[str] = None,
                 retry_on_submission_error: Optional[int] = None,
                 merge_sliced_step: bool = False,
                 pre_script: str = "",
                 post_script: str = "",
                 clean: bool = True,
                 remove_scheduling_strategies: bool = True,
                 envs: Optional[Dict[str, str]] = None,
                 merge_bohrium_job_group: bool = False,
                 terminate_grace_period: Optional[int] = None,
                 ) -> None:
        self.host = host
        self.queue_name = queue_name
        self.port = port
        self.username = username
        self.password = password
        self.private_key_file = private_key_file
        if image is None:
            image = config["dispatcher_image"]
        self.image = image
        if image_pull_policy is None:
            image_pull_policy = config["dispatcher_image_pull_policy"]
        self.image_pull_policy = image_pull_policy
        if isinstance(command, str):
            command = [command]
        self.command = command
        if isinstance(remote_command, str):
            remote_command = [remote_command]
        self.remote_command = remote_command
        self.map_tmp_dir = map_tmp_dir
        self.docker_executable = docker_executable
        self.singularity_executable = singularity_executable
        self.podman_executable = podman_executable
        if self.docker_executable is not None or \
                self.singularity_executable is not None or \
                self.podman_executable is not None:
            self.map_tmp_dir = False
        self.container_args = container_args
        self.work_root = "."
        self.remote_root = remote_root
        self.retry_on_submission_error = retry_on_submission_error
        self.merge_sliced_step = merge_sliced_step
        self.pre_script = pre_script
        self.post_script = post_script
        self.clean = clean
        self.remove_scheduling_strategies = remove_scheduling_strategies
        self.envs = envs
        self.merge_bohrium_job_group = merge_bohrium_job_group
        self.terminate_grace_period = terminate_grace_period

        conf = {}
        if json_file is not None:
            with open(json_file, "r") as f:
                conf = json.load(f)

        self.machine_dict = {
            "batch_type": "Slurm",
            "context_type": "SSHContext",
            "local_root": self.work_root,
            "remote_profile": {
                "hostname": self.host,
                "username": self.username,
                "port": self.port,
                "timeout": 10
            }
        }
        if self.password is not None:
            self.machine_dict["remote_profile"]["password"] = self.password
        if self.remote_root is not None:
            self.machine_dict["remote_root"] = self.remote_root
        else:
            if self.username == "root":
                self.machine_dict["remote_root"] = "/root/dflow/workflows"
            else:
                self.machine_dict["remote_root"] = "/home/%s/dflow/workflows"\
                    % self.username
        if self.private_key_file is not None:
            self.machine_dict["remote_profile"]["key_filename"] = \
                "./root/.ssh/" + os.path.basename(self.private_key_file)
        if "machine" in conf:
            update_dict(self.machine_dict, conf["machine"])
        if machine_dict is not None:
            update_dict(self.machine_dict, machine_dict)

        if self.machine_dict["context_type"] == "Bohrium":
            from . import bohrium
            if "batch_type" not in self.machine_dict:
                self.machine_dict["batch_type"] = "Bohrium"
            self.bohrium_ticket = bohrium.config["ticket"]
            if self.bohrium_ticket is not None:
                self.machine_dict["remote_profile"]["email"] = ""
            elif "email" not in self.machine_dict["remote_profile"]:
                if bohrium.config["username"] is not None:
                    self.machine_dict["remote_profile"]["email"] = \
                        bohrium.config["username"]
                else:
                    self.machine_dict["remote_profile"]["email"] = input(
                        "Bohrium email: ")
            if self.bohrium_ticket is not None:
                self.machine_dict["remote_profile"]["password"] = ""
            elif "password" not in self.machine_dict["remote_profile"]:
                if bohrium.config["password"] is not None:
                    self.machine_dict["remote_profile"]["password"] = \
                        bohrium.config["password"]
                else:
                    self.machine_dict["remote_profile"]["password"] = getpass(
                        "Bohrium password: ")
            if "program_id" not in self.machine_dict["remote_profile"]:
                if bohrium.config["project_id"] is not None:
                    self.machine_dict["remote_profile"]["program_id"] = int(
                        bohrium.config["project_id"])
                else:
                    self.machine_dict["remote_profile"]["program_id"] = int(
                        input("Bohrium program ID: "))
            if "input_data" not in self.machine_dict["remote_profile"]:
                self.machine_dict["remote_profile"]["input_data"] = {}
            input_data = self.machine_dict["remote_profile"]["input_data"]
            input_data["log_file"] = "log"
            if "job_type" not in input_data:
                input_data["job_type"] = "container"
            if "platform" not in input_data:
                input_data["platform"] = "ali"
            if "scass_type" not in input_data:
                input_data["scass_type"] = "c4_m8_cpu"
            if "job_name" not in input_data:
                input_data["job_name"] = "{{pod.name}}"
            if "output_log" not in input_data:
                input_data["output_log"] = True

        # set env to prevent dispatcher from considering different tasks as one
        self.resources_dict = {
            "number_node": 1,
            "cpu_per_node": 1,
            "group_size": 1,
            "envs": {
                "DFLOW_WORKFLOW": "{{workflow.name}}",
                "DFLOW_POD": "{{pod.name}}",
                "PYTHONUNBUFFERED": "true",
            }
        }
        if self.queue_name is not None:
            self.resources_dict["queue_name"] = self.queue_name
        if "resources" in conf:
            update_dict(self.resources_dict, conf["resources"])
        if resources_dict is not None:
            update_dict(self.resources_dict, resources_dict)

        self.task_dict = {
            "task_work_path": "./",
            "outlog": "log",
            "errlog": "log"
        }
        if task_dict is not None:
            update_dict(self.task_dict, task_dict)

    def modify_step(self, step):
        if not self.merge_bohrium_job_group:
            return
        assert self.machine_dict["context_type"] == "Bohrium"
        assert getattr(step.template, "slices", None) is not None
        if step.prepare_step is None:
            init_template = InitArtifactForSlices(
                template=step.template,
                image=step.util_image,
                command=step.util_command,
                image_pull_policy=step.util_image_pull_policy,
                key=step.key)
            step.prepare_step = step.__class__(
                name="%s-init-artifact" % step.name,
                template=init_template)
            step.prepare_step.inputs.parameters["dflow_group_key"] = \
                InputParameter(
                    value=re.sub("{{=?item.*}}", "group", str(step.key)))
        init_template = step.prepare_step.template
        init_template.outputs.parameters["bohr_job_group_id"] = \
            OutputParameter(
                value_from_path="%s/outputs/parameters/"
                "bohr_job_group_id" % init_template.tmp_root)
        step.prepare_step.outputs.parameters["bohr_job_group_id"] = \
            deepcopy(init_template.outputs.parameters[
                "bohr_job_group_id"])
        script = "import json, os\n"
        if step.key is not None:
            group_name = "{{{{workflow.name}}}}-"\
                "{{{{inputs.parameters.dflow_group_key}}}}"
        else:
            group_name = "{{{{pod.name}}}}"
        from . import bohrium
        add_url = "%s/brm/v1/job_group/add" % bohrium.config["bohrium_url"]
        data = '{{"name": "%s", "projectId": %s}}' % (
            group_name, self.machine_dict["remote_profile"]["program_id"])
        headers = "-H 'Content-type: application/json'"
        if self.bohrium_ticket is not None:
            headers += " -H 'Brm-Ticket: %s'" % self.bohrium_ticket
            cmd = "curl -X POST %s -d '%s' %s" % (headers, data, add_url)
            script += """ret = os.popen('''%s''').read()\n""" % cmd
        else:
            login_data = '{{"username": "%s", "password": "%s"}}' % (
                self.machine_dict["remote_profile"]["email"],
                self.machine_dict["remote_profile"]["password"])
            login_url = "%s/account/login" % bohrium.config["bohrium_url"]
            cmd = "curl -X POST -d '%s' %s" % (login_data, login_url)
            script += """ret = os.popen('''%s''').read()\n""" % cmd
            script += "print(ret)\n"
            script += "token = json.loads(ret)['data']['token']\n"
            headers += " -H 'Authorization: Bearer %s'"
            cmd = "curl -X POST %s -d '%s' %s" % (headers, data, add_url)
            script += """ret = os.popen('''%s''' %% token).read()\n""" % cmd
        script += "print(ret)\n"
        script += "group_id = json.loads(ret)['data']['groupId']\n"
        script += "os.makedirs('{tmp_root}/outputs/parameters', "\
            "exist_ok=True)\n"
        script += "with open('{tmp_root}/outputs/parameters/"\
            "bohr_job_group_id', 'w') as f:\n"
        script += "    f.write(str(group_id))\n"
        init_template.post_script = script
        init_template.render_script()
        step.inputs.parameters["bohr_job_group_id"] = InputParameter(
            value=step.prepare_step.outputs.parameters[
                "bohr_job_group_id"])

    def render(self, template):
        if not isinstance(template, ScriptOPTemplate):
            return template

        machine_dict = deepcopy(self.machine_dict)
        task_dict = deepcopy(self.task_dict)
        new_template = deepcopy(template)
        new_template.name += "-" + randstr()
        new_template.image = self.image
        new_template.image_pull_policy = self.image_pull_policy
        new_template.command = self.command
        if self.remove_scheduling_strategies:
            new_template.node_selector = None
            new_template.tolerations = []
            new_template.affinity = None
            # dispatcher pod use ephemeral storage currently
            if new_template.requests is not None:
                new_template.requests = {
                    k: v for k, v in new_template.requests.items()
                    if k == "ephemeral-storage"}
            if new_template.limits is not None:
                new_template.limits = {
                    k: v for k, v in new_template.limits.items()
                    if k == "ephemeral-storage"}

        remote_command = template.command if self.remote_command is None \
            else self.remote_command
        cmd = ""
        if self.map_tmp_dir:
            cmd += "sed -i \"s#\\$(pwd)#$(pwd)#g\" script && "

        cmd += run_script(template.image, remote_command,
                          self.docker_executable, self.singularity_executable,
                          self.podman_executable, args=self.container_args)
        task_dict["command"] = cmd
        task_dict["forward_files"] = ["script"]
        for art in template.inputs.artifacts.values():
            task_dict["forward_files"].append("./" + art.path)
        for par in template.inputs.parameters.values():
            if par.save_as_artifact:
                task_dict["forward_files"].append("./" + par.path)
        merge = self.merge_sliced_step and hasattr(template, "slices") and\
            template.slices is not None
        if merge:
            assert not template.slices.sub_path, "sub_path mode of slices "\
                "is incompatible with merge_sliced_step"
            assert template.slices.group_size is None, "group_size of slices "\
                "is incompatible with merge_sliced_step"
            assert template.slices.pool_size is None, "pool_size of slices "\
                "is incompatible with merge_sliced_step"
            sliced_output_parameters = template.slices.output_parameter.copy()
            if "dflow_success_tag" in template.outputs.parameters:
                sliced_output_parameters.append("dflow_success_tag")
        task_dict["backward_files"] = []
        for art in template.outputs.artifacts.values():
            task_dict["backward_files"].append("./" + art.path)
        for name, par in template.outputs.parameters.items():
            if par.save_as_artifact:
                task_dict["backward_files"].append(
                    "./" + par.value_from_path)
            elif par.value_from_path is not None and not (
                    merge and sliced_output_parameters):
                task_dict["backward_files"].append(
                    "./" + par.value_from_path)

        new_template.script = self.pre_script
        new_template.script += "import os\n"
        # compatible for debug mode
        new_template.script += "if not os.getcwd().endswith('workdir'):\n"
        new_template.script += "    os.chdir('/')\n"
        if any([art.save_as_parameter for art in
                template.inputs.artifacts.values()]) and \
                isinstance(template, PythonOPTemplate):
            template = deepcopy(template)
            if machine_dict["context_type"] == "Bohrium":
                template.download_method = "bohrium_download"
            elif machine_dict["context_type"] == "SSHContext":
                template.download_method = "remote_download"
            template.render_script()
            new_template.volumes = [v for v in new_template.volumes
                                    if v.name not in ["launching", "dev-fuse"]]
            new_template.mounts = [m for m in new_template.mounts
                                   if m.name != "launching"]
            new_template.sidecars = [s for s in new_template.sidecars
                                     if not s.name.startswith("rclone-")]
            new_template.script += handle_packages_script(
                "./tmp/inputs/artifacts/dflow_python_packages")
            new_template.script += "from dflow import config\n"
            new_template.script += "config['artifact_register'] = %s\n" % \
                json.dumps(config["artifact_register"])
        new_template.script += "with open('script', 'w') as f:\n"
        new_template.script += "    f.write(r\"\"\"\n"
        if self.map_tmp_dir:
            script = render_script_with_tmp_root(template, "$(pwd)/tmp")
        else:
            script = template.script
        # keep the script human-readable
        script = script.replace("\"\"\"", "'''")
        new_template.script += script
        new_template.script += "\"\"\")\n"

        if machine_dict["context_type"] == "Bohrium":
            if self.bohrium_ticket is not None:
                if new_template.envs is None:
                    new_template.envs = {}
                new_template.envs["BOHR_TICKET"] = self.bohrium_ticket
            if "image_name" not in machine_dict["remote_profile"][
                    "input_data"]:
                machine_dict["remote_profile"]["input_data"][
                    "image_name"] = template.image

        machine_dict["local_root"] = self.work_root
        new_template.script += "import json, shlex\n"
        if config["dispatcher_debug"]:
            new_template.script += "import logging\n"
            new_template.script += "from dpdispatcher.dlog import dlog\n"
            new_template.script += "dlog.setLevel(logging.DEBUG)\n"
        new_template.script += "from dpdispatcher import Machine, Resources,"\
            " Task, Submission\n"
        machine_json = json.dumps(machine_dict)
        if self.merge_bohrium_job_group:
            new_template.inputs.parameters["bohr_job_group_id"] = \
                InputParameter()
            machine_dict["remote_profile"]["input_data"]["bohr_job_group_id"]\
                = "{{inputs.parameters.bohr_job_group_id}}"
            machine_json = json.dumps(machine_dict)
            machine_json = machine_json.replace(
                "\"{{inputs.parameters.bohr_job_group_id}}\"",
                "{{inputs.parameters.bohr_job_group_id}}")
        new_template.script += "machine = Machine.load_from_dict(json.loads("\
            "r'%s'))\n" % machine_json
        for name, art in template.inputs.artifacts.items():
            if machine_dict["context_type"] == "Bohrium" and \
                    art.save_as_parameter:
                new_template.script += "from dflow import CustomArtifact\n"
                new_template.script += "CustomArtifact.from_urn"\
                    "('{{inputs.parameters.dflow_art_%s}}').modify_config("\
                    "'%s', machine)\n" % (name, name)

        new_template.script += "resources = Resources.load_from_dict(json."\
            "loads(r'%s'))\n" % json.dumps(self.resources_dict)
        if new_template.envs is not None:
            for k in new_template.envs.keys():
                new_template.script += "resources.envs['%s'] = "\
                    "os.environ.get('%s')\n" % (k, k)
        new_template.script += "resources.envs['ARGO_TEMPLATE'] = "\
            "shlex.quote(os.environ.get('ARGO_TEMPLATE'))\n"
        new_template.script += "task = Task.load_from_dict(json.loads(r'%s'))"\
            "\n" % json.dumps(task_dict)
        new_template.script += "task.forward_files = list(filter("\
            "os.path.exists, task.forward_files))\n"
        new_template.script += "for f in task.backward_files:\n"
        new_template.script += "    os.makedirs(os.path.dirname(f), "\
            "exist_ok=True)\n"
        if merge:
            new_template.inputs.parameters["dflow_with_param"] = \
                InputParameter(value="")
            new_template.inputs.parameters["dflow_sequence_start"] = \
                InputParameter(value=0)
            new_template.inputs.parameters["dflow_sequence_end"] = \
                InputParameter(value=None)
            new_template.inputs.parameters["dflow_sequence_count"] = \
                InputParameter(value=None)
            new_template.inputs.parameters["dflow_sequence_format"] = \
                InputParameter(value="")
            new_template.script += "from copy import deepcopy\n"
            new_template.script += "with open('script', 'r') as f:\n"
            new_template.script += "    script = f.read()\n"
            new_template.script += "tasks = []\n"
            new_template.script += "with_param = r'''{{inputs.parameters."\
                "dflow_with_param}}'''\n"
            new_template.script += "if with_param != '':\n"
            new_template.script += "    item_list = json.loads(with_param)\n"
            new_template.script += "else:\n"
            new_template.script += "    start = json.loads('{{inputs."\
                "parameters.dflow_sequence_start}}')\n"
            new_template.script += "    count = json.loads('{{inputs."\
                "parameters.dflow_sequence_count}}')\n"
            new_template.script += "    end = json.loads('{{inputs."\
                "parameters.dflow_sequence_end}}')\n"
            new_template.script += "    format = '{{inputs.parameters."\
                "dflow_sequence_format}}'\n"
            new_template.script += "    if count is not None:\n"
            new_template.script += "        r = range(start, start + count)\n"
            new_template.script += "    elif end is not None:\n"
            new_template.script += "        if end > start:\n"
            new_template.script += "            r = range(start, end + 1)\n"
            new_template.script += "        else:\n"
            new_template.script += "            r = range(start, end - 1, -1)"\
                "\n"
            new_template.script += "    item_list = [format % i if format != "\
                "'' else i for i in r]\n"
            new_template.script += "for i, item in enumerate(item_list):\n"
            new_template.script += "    new_task_dict = deepcopy(task."\
                "serialize())\n"
            new_template.script += "    new_script = script\n"
            for k, v in new_template.dflow_vars.items():
                if "item" in k:
                    old = "'{{inputs.parameters.%s}}'" % v
                    new = "item" if k == "{{item}}" else "item[%s]" % k[6:-2]
                    new_template.script += "    new_script = new_script."\
                        "replace(%s, %s if isinstance(%s, str) else "\
                        "json.dumps(%s))\n" % (old, new, new, new)
            for name in template.slices.output_parameter:
                new_template.script += "    new_script = new_script.replace("\
                    "\"handle_output_parameter('%s'\", "\
                    "\"handle_output_parameter('%s_\" + str(i) + \"'\")\n" % (
                        name, name)
                path = template.outputs.parameters[name].value_from_path
                new_template.script += "    os.makedirs(os.path.dirname("\
                    "'./%s'), exist_ok=True)\n" % path
                new_template.script += "    new_task_dict['backward_files']"\
                    ".append('./%s_' + str(i))\n" % path
            if "dflow_success_tag" in template.outputs.parameters:
                new_template.script += "    new_script = new_script.replace("\
                    "'success_tag', 'success_tag_' + str(i))\n"
                path = template.outputs.parameters[
                    "dflow_success_tag"].value_from_path
                new_template.script += "    os.makedirs(os.path.dirname("\
                    "'./%s'), exist_ok=True)\n" % path
                new_template.script += "    new_task_dict['backward_files']"\
                    ".append('./%s_' + str(i))\n" % path
            new_template.script += "    new_task_dict['backward_files']"\
                ".append('log')\n"  # work around no Bohrium result file
            new_template.script += "    with open('script' + str(i), 'w')"\
                " as f:\n"
            new_template.script += "        f.write(new_script)\n"
            new_template.script += "    new_task_dict['command'] = new_task_"\
                "dict['command'].replace('script', 'script' + str(i))\n"
            new_template.script += "    new_task_dict['forward_files'][0] = "\
                "'script' + str(i)\n"
            new_template.script += "    tasks.append(Task.load_from_dict("\
                "new_task_dict))\n"
            new_template.script += "submission = Submission(work_base='.', "\
                "machine=machine, resources=resources, task_list=tasks)\n"
        else:
            new_template.script += "submission = Submission(work_base='.', "\
                "machine=machine, resources=resources, task_list=[task])\n"
        new_template.script += "def sigterm_handler(signum, frame):\n"
        new_template.script += "    print('Got SIGTERM, kill unfinished tasks"\
            "!')\n"
        new_template.script += "    submission.remove_unfinished_tasks()\n"
        if self.terminate_grace_period:
            new_template.script += "    import time\n"
            new_template.script += "        time.sleep(%s)\n" % \
                self.terminate_grace_period
        new_template.script += "import signal\n"
        new_template.script += "signal.signal(signal.SIGTERM, sigterm_handler"\
            ")\n"
        if self.retry_on_submission_error:
            new_template.script += "from dpdispatcher.dpcloudserver.client "\
                "import RequestInfoException\n"
            new_template.script += "for retry in range(%s):\n" % \
                self.retry_on_submission_error
            new_template.script += "    try:\n"
            new_template.script += "        print('retry ' + str(retry))\n"
            new_template.script += "        submission.run_submission(clean="\
                "%s)\n" % self.clean
            new_template.script += "        break\n"
            new_template.script += "    except RequestInfoException:\n"
            new_template.script += "        import traceback\n"
            new_template.script += "        traceback.print_exc()\n"
            new_template.script += "        import time\n"
            new_template.script += "        time.sleep(2**retry)\n"
        else:
            new_template.script += "submission.run_submission(clean=%s)\n" % \
                self.clean
        if merge:
            for name in sliced_output_parameters:
                path = template.outputs.parameters[name].value_from_path
                new_template.script += "res = []\n"
                new_template.script += "for i in range(len(item_list)):\n"
                new_template.script += "    fname = './%s_' + str(i)\n" % path
                new_template.script += "    if os.path.isfile(fname):\n"
                new_template.script += "        with open(fname, 'r') as f:\n"
                new_template.script += "            res.append(f.read())\n"
                new_template.script += "with open('./%s', 'w') as f:\n" % path
                new_template.script += "    f.write(json.dumps(res))\n"
        new_template.script_rendered = True

        if machine_dict["context_type"] == "Bohrium":
            new_template.script += "for job in submission.belonging_jobs:\n" \
                "    assert machine.get_exit_code(job) == 0\n"
        else:
            # workaround for unavailable exit code of dispatcher
            # check output files explicitly
            for art in template.outputs.artifacts.values():
                new_template.script += "assert os.path.exists('./%s')\n" % \
                    art.path
            for par in template.outputs.parameters.values():
                if par.save_as_artifact or (par.value_from_path is not None and
                                            not hasattr(par, "default")):
                    new_template.script += "assert os.path.exists('./%s')\n" %\
                        par.value_from_path

        new_template.script += self.post_script
        if self.private_key_file is not None:
            key = upload_s3(self.private_key_file)
            private_key_artifact = S3Artifact(key=key)
            new_template.inputs.artifacts["dflow_private_key"] = InputArtifact(
                path="/root/.ssh/" + os.path.basename(self.private_key_file),
                source=private_key_artifact, mode=0o600)
        elif config["private_key_host_path"] is not None:
            new_template.volumes.append(V1Volume(
                name="dflow-private-key", host_path=V1HostPathVolumeSource(
                    path=config["private_key_host_path"])))
            new_template.mounts.append(V1VolumeMount(
                name="dflow-private-key", mount_path="/root/.ssh"))

        if self.envs is not None:
            if new_template.envs is None:
                new_template.envs = {}
            new_template.envs.update(self.envs)
        return new_template


class DispatcherArtifact(CustomArtifact, ABC):
    @abc.abstractmethod
    def modify_config(self, name: str, machine) -> str:
        pass

    @abc.abstractmethod
    def remote_download(self, name: str, path: str):
        pass

    @abc.abstractmethod
    def bohrium_download(self, name: str, path: str):
        pass
